############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

import zmq

from hermes.utils.zmq_utils import CMD_HELLO

from hermes.base.nodes.node_interface import NodeInterface
from hermes.base.state_interface import StateInterface


class AbstractNodeState(StateInterface):
    """Abstract class for the Node FSM.

    Can be externally triggered into the KILL state from any child State class.
    """

    def __init__(self, context: NodeInterface):
        """Constructor of the AbstractNodeState parent class.

        Args:
            context (NodeInterface): Reference to the Node object.
        """
        self._context = context

    def is_continue(self) -> bool:
        return True

    def kill(self) -> None:
        self._context._set_state(KillState(self._context))


class StartState(AbstractNodeState):
    """Starting state of the Node that initializes corresponding internal logic of the specified Node.

    Activates data poller in case Node goes into KillState.
    """

    def run(self):
        self._context._initialize()
        self._context._activate_data_poller()
        self._context._set_state(SyncState(self._context))


class SyncState(AbstractNodeState):
    """Synchronization state of the Node to coordinate readiness to start working to the local master Broker."""

    def __init__(self, context: NodeInterface):
        super().__init__(context)
        self._sync = context._get_sync_socket()

    def run(self):
        self._sync.send_multipart(
            [self._context._log_source_tag().encode("utf-8"), CMD_HELLO.encode("utf-8")]
        )
        host, cmd = self._sync.recv_multipart()
        print(
            "%s received %s from %s."
            % (
                self._context._log_source_tag(),
                cmd.decode("utf-8"),
                host.decode("utf-8"),
            ),
            flush=True,
        )
        self._context._set_state(RunningState(self._context))


class RunningState(AbstractNodeState):
    """Running state of the Node with the steady-state main loop listening for data and triggering user-provided callbacks on it."""

    def __init__(self, context: NodeInterface):
        super().__init__(context)
        self._context._activate_kill_poller()
        self._context._on_sync_complete()

    def run(self):
        poll_res: tuple[list[zmq.SyncSocket], list[int]] = self._context._poll()
        self._context._on_poll(poll_res)


class KillState(AbstractNodeState):
    """Termination state of the Node that gracefully processes the external KILL signal from the local master Broker."""

    def run(self):
        self._context._deactivate_kill_poller()
        self._context._send_kill_to_broker()
        self._context._trigger_stop()
        self._context._set_state(JoinState(self._context))

    def kill(self):
        pass


class JoinState(AbstractNodeState):
    """Gracefully exit state of the Node to coordinate closure to the local master Broker."""

    def run(self):
        poll_res: tuple[list[zmq.SyncSocket], list[int]] = self._context._poll()
        self._context._on_poll(poll_res)

    def is_continue(self):
        return not self._context._is_done

    def kill(self):
        pass



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from multiprocessing import Event, set_start_method
from multiprocessing.synchronize import Event as _EventClass
import subprocess
import threading
import os
import sys
import yaml
import json
import argparse
import platform

from hermes.__version__ import __version__
from hermes.base.broker.broker import Broker
from hermes.utils.argparse_utils import ParseExperimentKwargs, validate_path
from hermes.utils.time_utils import get_ref_time, get_time
from hermes.utils.zmq_utils import (
    PORT_BACKEND,
    PORT_FRONTEND,
    PORT_KILL,
    PORT_SYNC_HOST,
)
from hermes.utils.types import LoggingSpec, VideoCodec, AudioCodec, VideoFormatEnum


# TODO: replace with HERMES-branded font
HERMES = r"""
______  ________________________  ___________________
___  / / /__  ____/__  __ \__   |/  /__  ____/_  ___/
__  /_/ /__  __/  __  /_/ /_  /|_/ /__  __/  _____ \ 
_  __  / _  /___  _  _, _/_  /  / / _  /___  ____/ / 
/_/ /_/  /_____/  /_/ |_| /_/  /_/  /_____/  /____/  
                                                     
"""
DESCRIPTION = (
    HERMES + "Heterogeneous edge realtime measurement and execution system "
    "for continual multimodal data acquisition and AI processing."
)
EPILOG = (
    "Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab.\n"
    "Created 2024-2025 at KU Leuven for the AidWear, AID-FOG, and RevalExo "
    "projects of prof. Bart Vanrumste, by Maxim Yudayev [https://yudayev.com]."
)


def define_parser() -> argparse.ArgumentParser:
    """Create and return the CLI argument parser for the HERMES application.

    Returns:
        argparse.ArgumentParser: Configured argument parser ready to parse
            command-line arguments for the application.
    """
    parser = argparse.ArgumentParser(
        prog="HERMES",
        description=DESCRIPTION,
        epilog=EPILOG,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="increase level of logging verbosity [0,3]",
    )
    parser.add_argument(
        "--version", action="version", version="%(prog)s v" + __version__
    )

    parser.add_argument(
        "--out_dir",
        "-o",
        type=validate_path,
        dest="out_dir",
        required=True,
        help="path to the output directory of the current host device",
    )
    parser.add_argument(
        "--experiment",
        "-e",
        nargs="*",
        action=ParseExperimentKwargs,
        help="key-value pair tags detailing the experiment, used for "
        "directory creation and metadata on files",
    )
    parser.add_argument(
        "--time",
        "-t",
        type=float,
        dest="log_time_s",
        default=get_time(),
        help="master start time of the system",
    )
    parser.add_argument(
        "--duration",
        "-d",
        type=int,
        dest="duration_s",
        default=None,
        help="duration in seconds, if using for recording only (to be used only by master)",
    )
    parser.add_argument(
        "--config_file",
        "-f",
        type=validate_path,
        default=None,
        help="path to the configuration file for the current host device, "
        "instead of the CLI arguments",
    )
    parser.add_argument(
        "--json",
        "-j",
        default=None,
        help="serialized JSON configuration for slave host setup by master over SSH",
    )

    return parser


def parse_args(parser: argparse.ArgumentParser) -> argparse.Namespace:
    """Parse CLI arguments and apply configuration file overrides.

    This function wraps `parser.parse_args()` to allow optional provision
    of configuration from a YAML config file (see `parse_config_file`)
    or from a JSON string (see `parse_json_string`),
    and to load codec specifications if required (see `load_codec_spec`).
    It also validates the parsed configuration for completeness of required fields.

    Args:
        parser (argparse.ArgumentParser): The parser created by `define_parser()`.

    Returns:
        argparse.Namespace: The fully-resolved CLI arguments namespace.

    Raises:
        Exception: If neither `--config_file` nor `--json` is provided.
    """
    args = parser.parse_args()

    if args.config_file is not None:
        parser, args = parse_config_file(parser, args)
    elif args.json is not None:
        parser, args = parse_json_string(parser, args)
    else:
        raise Exception("Either `--config_file` or `--json` must be provided.", "Got: \n", vars(args))

    # TODO: use pydantic to validate args here

    args = load_codec_spec(args)
    print(HERMES, flush=True)
    return args


def parse_config_file(
    parser: argparse.ArgumentParser, args: argparse.Namespace
) -> tuple[argparse.ArgumentParser, argparse.Namespace]:
    """Override parser defaults with values from a YAML config file.

    If `args.config_file` is set, this function reads the file, performs
    environment variable substitution for patterns like `${VAR}`, and uses
    the resulting YAML mapping to set default values on the provided
    `parser`. After applying defaults the parser is re-parsed so the
    returned `args` reflect any config-file-provided values.

    Args:
        parser (argparse.ArgumentParser): The argument parser to update.
        args (argparse.Namespace): The namespace produced by the initial
            `parse_args()` call.

    Returns:
        tuple[argparse.ArgumentParser, argparse.Namespace]: The parser (with
            updated defaults) and the re-parsed args namespace.

    Exits:
        Calls `exit()` with an error message if the YAML cannot be parsed.
    """
    with open(args.config_file, "r") as f:
        try:
            config_str = f.read()
            config_str = inject_env_vars(config_str)
            config: dict = yaml.safe_load(config_str)
            parser.set_defaults(**config)
        except yaml.YAMLError as e:
            print(e, flush=True)
            exit("Error parsing YAML file.")
    args = parser.parse_args()
    return parser, args


def parse_json_string(
    parser: argparse.ArgumentParser, args: argparse.Namespace
) -> tuple[argparse.ArgumentParser, argparse.Namespace]:
    """Override parser defaults with values from a JSON config string.

    If `args.json` is set, this function reads the JSON string, performs
    environment variable substitution for patterns like `${VAR}`, and uses
    the resulting JSON mapping to set default values on the provided
    `parser`. After applying defaults the parser is re-parsed so the
    returned `args` reflect any config-file-provided values.

    Args:
        parser (argparse.ArgumentParser): The argument parser to update.
        args (argparse.Namespace): The namespace produced by the initial
            `parse_args()` call.

    Returns:
        tuple[argparse.ArgumentParser, argparse.Namespace]: The parser (with
            updated defaults) and the re-parsed args namespace.

    Exits:
        Calls `exit()` with an error message if the JSON cannot be parsed.
    """
    try:
        config_str = args.json
        config_str = inject_env_vars(config_str)
        if platform.system() == 'Windows':
            config_str = config_str.replace("'", '"')
        config: dict = json.loads(config_str)
        parser.set_defaults(**config)
    except json.JSONDecodeError as e:
        print(e, flush=True)
        exit("Error parsing JSON string.")
    args = parser.parse_args()
    return parser, args


def inject_env_vars(config_str: str) -> str:
    """Replace environment variable placeholders with corresponding values.

    This function searches for patterns like `${VAR}` in the provided configuration
    string and replaces them with the corresponding environment variable values.

    Args:
        config_str (str): The configuration string potentially containing
            environment variable patterns.

    Returns:
        str: The same configuration string with environment variables injected in place of placeholders.
    """
    for key, value in os.environ.items():
        config_str = config_str.replace(f"${{{key}}}", value)
    return config_str


def replace_video_format_nested(config: dict) -> dict:
    """Recursively replace `video_image_format` strings with enum values.

    Args:
        config (dict): A node specification dictionary potentially containing the key.
    """
    if "video_image_format" in config:
        config["video_image_format"] = VideoFormatEnum[config["video_image_format"]]
        return config

    for key, value in config.items():
        if isinstance(value, dict):
            config[key] = replace_video_format_nested(value)

    return config


def load_codec_spec(args: argparse.Namespace) -> argparse.Namespace:
    """Load codec configuration YAML files into the `args.logging_spec`.

    When video/audio streaming is enabled in `args.logging_spec`, this
    function opens the corresponding codec configuration files, parses the
    YAML and constructs `VideoCodec` / `AudioCodec` objects which are
    attached back to `args.logging_spec` under the keys
    `'video_codec'` and `'audio_codec'` respectively.

    Args:
        args (argparse.Namespace): Parsed CLI arguments containing a
            `logging_spec` mapping with codec filepaths.

    Returns:
        argparse.Namespace: The same `args` object with codec objects
            injected into `args.logging_spec` when applicable.
    """
    if "stream_video" in args.logging_spec and args.logging_spec["stream_video"]:
        with open(args.video_codec_config_filepath, "r") as f:
            try:
                args.logging_spec["video_codec"] = VideoCodec(**yaml.safe_load(f))
            except yaml.YAMLError as e:
                print(e)
            # Replace `video_image_format` with appropriate enum value
            args.producer_specs = [
                replace_video_format_nested(spec) for spec in args.producer_specs
            ]
    if "stream_audio" in args.logging_spec and args.logging_spec["stream_audio"]:
        with open(args.audio_codec_config_filepath, "r") as f:
            try:
                args.logging_spec["audio_codec"] = AudioCodec(**yaml.safe_load(f))
            except yaml.YAMLError as e:
                print(e)
    return args


def init_output_files(args: argparse.Namespace) -> tuple[float, str, str]:
    """Prepare output directories and return paths for logging files.

    This function computes the experiment `log_dir` using `args.out_dir`
    and `args.experiment`, creates the directory structure and returns the
    resolved master logging time, the log directory path and a path for a
    host-specific log history file.

    Args:
        args (argparse.Namespace): Parsed CLI arguments containing
            `out_dir`, `experiment` and `host_ip`.

    Returns:
        tuple[float, str, str]: A tuple containing `(log_time_s, log_dir,
            log_history_filepath)` where `log_time_s` is the master start time,
            `log_dir` is the directory created for logs and
            `log_history_filepath` is the per-host log filename.

    Exits:
        Calls `exit()` if the experiment directory already exists.
    """
    log_time_s = args.log_time_s if args.log_time_s is not None else get_time()
    log_dir: str = os.path.join(
        args.out_dir, *map(lambda tup: "_".join(tup), args.experiment.items())
    )
    log_history_filepath: str = os.path.join(log_dir, "%s.log" % args.host_ip)

    try:
        os.makedirs(log_dir)
    except OSError:
        exit(
            "'%s' already exists. Update experiment YML file with correct data for this subject."
            % log_dir
        )

    return log_time_s, log_dir, log_history_filepath


def configure_specs(
    args: argparse.Namespace, log_time_s: float, log_dir: str
) -> tuple[argparse.Namespace, list[dict]]:
    """Build logging specification and inject settings into node specs.

    Constructs a `LoggingSpec` object from provided arguments and updates
    each node spec in `args.producer_specs`, `args.consumer_specs` and
    `args.pipeline_specs` by populating the `settings` mapping with host
    information, ports and the `logging_spec` object. This prepares node
    specs for broker initialization.

    Args:
        args (argparse.Namespace): Parsed CLI arguments containing node
            spec lists and host information.
        log_time_s (float): Master logging start time.
        log_dir (str): Path to the directory where logs should be stored.

    Returns:
        tuple[argparse.Namespace, list[dict], float]: The (possibly unchanged)
            args object, a flat list of node spec dictionaries ready to be
            consumed by the `Broker`, host device's reference time for performance
            counters.
    """
    ref_time_s = get_ref_time()
    logging_spec = LoggingSpec(
        log_dir=log_dir,
        log_time_s=log_time_s,
        ref_time_s=ref_time_s,
        experiment=args.experiment,
        **args.logging_spec,
    )

    node_specs: list[dict] = (
        args.producer_specs + args.consumer_specs + args.pipeline_specs
    )
    for spec in node_specs:
        spec["settings"]["host_ip"] = args.host_ip
        spec["settings"]["logging_spec"] = logging_spec
        spec["settings"]["port_pub"] = PORT_BACKEND
        spec["settings"]["port_sub"] = PORT_FRONTEND
        spec["settings"]["port_sync"] = PORT_SYNC_HOST
        spec["settings"]["port_killsig"] = PORT_KILL

    return args, node_specs, ref_time_s


def parse_stdin(
    broker: Broker,
    is_master: bool,
    is_ready_event: _EventClass,
    is_done_event: _EventClass,
    is_quit_event: _EventClass
) -> None:
    """Parse user keyboard inputs into a fanout queue for HERMES nodes.

    Blocking stdin capture loop as a daemon, that fans out keyboard inputs
    to Broker's subprocesses and all the Nodes.
    On Windows, daemon threads get cleaned up automatically on Python interpreter exit.
    On Linux/Mac, uses select on the system stdin to avoid blocking indefinitely. 

    Args:
        broker (Broker): Host's only HERMES instance. 
        is_master (bool): Whether the Broker is the master in the setup.
        is_ready_event (Event): Synchronization flag whether the Broker and its subprocesses finished setting up.
        is_done_event (Event): Synchronization flag whether the Broker is done and gracefully wrapped up.
        is_quit_event (Event): Synchronization flag to indicate to the Broker to start the graceful quit procedure.
    """
    user_input = ""
    termination_char = "Q"
    # is_ready_event.wait()  # NOTE: deadlocks if a Node requires user input during the bring-up process.
    if platform.system() == 'Windows':
        while not is_done_event.is_set():
            user_input = input(">> ")
            if is_master and user_input == termination_char:
                is_quit_event.set()
            elif len(user_input):
                broker._fanout_user_input((get_time(), user_input))
    else:
        import select
        while not is_done_event.is_set():
            ready, _, _ = select.select([sys.stdin], [], [], 5.0)
            if ready:
                user_input = sys.stdin.readline().strip()
                if is_master and user_input == termination_char:
                    is_quit_event.set()
                elif len(user_input):
                    broker._fanout_user_input((get_time(), user_input))


def launch_slave_hosts(
    connections: list[dict], log_time_s: float, experiment: dict[str, str]
) -> list[subprocess.Popen]:
    """Launch slave HERMES hosts over SSH, each in a new interactive terminal window.

    This function constructs SSH commands to connect to each slave host
    specified in `connections` of the master's YAML config file.
    It parses slave config YAML files of each listed host, constructs the corresponding command that
    activates the remote host's Python virtual environment on each slave host,
    and runs `hermes-cli` with appropriate arguments in an interactive shell to accept user keyboard inputs.

    Args:
        connections (list[dict]): List of connection specifications for
            each slave host, including SSH credentials and paths.
        log_time_s (float): Master logging start time to pass to slaves.
        experiment (dict[str, str]): Experiment key-value pairs to pass to slaves.

    Returns:
        list[subprocess.Popen]: List of subprocess handles for the launched slave hosts.
    """

    experiment_str = ' '.join([f"{k}={v}" for k,v in experiment.items()])
    cmds = []
    for conn in connections:
        with open(conn["config_filepath"], "r") as f:
            try:
                file_buf = f.read()
                config: dict = yaml.safe_load(file_buf)
                config_str = json.dumps(config)

                if conn["platform"] == "Windows":
                    escaped_config_str = config_str.replace('"', "'")
                    remote_cmd = (
                        f"title HERMES - {conn['ssh_username']}@{conn['ssh_host_ip']} && "
                        f"cd /d {conn['project_dir']} && "
                        f"call .venv\\Scripts\\activate.bat && "
                        f"hermes-cli -o {conn['output_dir']} -t {log_time_s} -e {experiment_str} -j \"{escaped_config_str}\" && "
                        f"exit"
                    )
                else:
                    remote_cmd = (
                        f"echo -ne \"\033]0;HERMES - {conn['ssh_username']}@{conn['ssh_host_ip']}\007\" && "
                        f"source ~/.bash_profile 2>/dev/null || source ~/.profile 2>/dev/null; "
                        f"cd {conn['project_dir']} && "
                        f"source .venv/bin/activate && "
                        f"export PYTHONPATH=\"$(pwd):$PYTHONPATH\" && "
                        f"hermes-cli -o {conn['output_dir']} -t {log_time_s} -e {experiment_str} -j '{config_str}' && "
                        f"exit"
                    )

                cmds.append([
                    f"{conn['ssh_username']}@{conn['ssh_host_ip']}",
                    remote_cmd
                ])
            except yaml.YAMLError as e:
                print(e, flush=True)
                exit("Error parsing slave YAML files.")

    if platform.system() == "Windows":
        prog = ["cmd", "/c"]
    elif platform.system() == "Linux":
        prog = ["gnome-terminal", "--"]
    elif platform.system() == "Darwin":
        prog = ["open", "-a", "Terminal"]

    procs = []
    for cmd in cmds:
        procs.append(
            subprocess.Popen([
                *prog,
                "ssh", "-tt",
                "-o", "TCPKeepAlive=no",
                "-o", "ServerAliveInterval=30",
                cmd[0],
                cmd[1],
            ], creationflags=subprocess.CREATE_NEW_CONSOLE if platform.system() == "Windows" else 0)
        )
    return procs


def app():
    """Main entry point for the HERMES CLI application.

    This function wires together argument parsing, output directory
    creation, node specification configuration, and broker lifecycle
    management. When executed as the master broker it spawns the broker
    in a background thread and listens for a terminal 'Q' input to
    gracefully terminate the experiment.
    """
    parser = define_parser()
    args = parse_args(parser)

    log_time_s, log_dir, log_history_filepath = init_output_files(args)
    args, node_specs, ref_time_s = configure_specs(args, log_time_s, log_dir)

    set_start_method("spawn")

    # Launch slave hosts over SSH if the current broker is master and any connections are specified.
    if args.is_master_broker and args.connections:
        if platform.system() == "Windows":
            os.system(f"title HERMES - {args.host_ip}")
        else:
            os.system(f'echo -ne "\033]0;HERMES - {args.host_ip}\007"')
        slave_procs = launch_slave_hosts(args.connections, log_time_s, args.experiment)

    is_ready_event = Event()
    is_quit_event = Event()
    is_done_event = Event()

    # Create the broker and manage all the components of the experiment.
    local_broker: Broker = Broker(
        host_ip=args.host_ip,
        node_specs=node_specs,
        is_ready_event=is_ready_event,
        is_quit_event=is_quit_event,
        is_done_event=is_done_event,
        is_master_broker=args.is_master_broker,
    )

    # Connect broker to remote publishers at the wearable PC to get data from the wearable sensors.
    for ip in args.remote_publisher_ips:
        local_broker.connect_to_remote_broker(addr=ip)

    # Expose local wearable data to remote subscribers (e.g. edge server).
    if args.remote_subscriber_ips:
        local_broker.expose_to_remote_broker(args.remote_subscriber_ips)

    # Subscribe to the KILL signal of a remote machine.
    if args.is_remote_kill:
        local_broker.subscribe_to_killsig(addr=args.remote_kill_ip)

    stdin_thread = threading.Thread(
        target=parse_stdin,
        args=(
            local_broker,
            args.is_master_broker,
            is_ready_event,
            is_done_event,
            is_quit_event,
        ),
        daemon=True
    )
    stdin_thread.start()

    # Only master host runs with duration, others wait for commands.
    if args.is_master_broker:
        local_broker(args.duration_s)
    else:
        local_broker()

    if args.is_master_broker and args.connections:
        for proc in slave_procs: proc.wait()


if __name__ == "__main__":
    app()



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

import msgpack
import numpy as np


def encode_ndarray(obj: object) -> object:
    """Encodes NumPy contents of the provided object into bytes.

    Args:
        obj (object): NumPy array to encode as serializeable key-value pair.

    Returns:
        object: Encoded key-value mapping with NumPy arrays serialized to bytes.
    """
    if isinstance(obj, np.ndarray):
        return {
            "__numpy__": True,
            "shape": obj.shape,
            "dtype": str(obj.dtype),
            "bytes": obj.tobytes(),
        }
    return obj


def decode_ndarray(obj: object) -> object:
    """Decodes received bytes and reconstructs detected NumPy arrays into original objects.

    Args:
        obj (object): Raw bytes to convert into NumPy arrays if any detected.

    Returns:
        object: Decoded key-value mapping with NumPy arrays reconstructed.
    """
    if "__numpy__" in obj:
        obj = np.frombuffer(obj["bytes"], dtype=obj["dtype"]).reshape(obj["shape"])
    return obj


def convert_bytes_keys_to_strings(obj: object) -> object:
    """Recursively decodes keys of the key-value mappings into strings.

    Args:
        obj (object): Object whose keys to decode.

    Returns:
        object: Processed object with keys converted into proper dictionary string fields.
    """
    if isinstance(obj, dict):
        return {
            (
                key.decode("utf-8") if isinstance(key, bytes) else key
            ): convert_bytes_keys_to_strings(value)
            for key, value in obj.items()
        }
    elif isinstance(obj, list):
        return [convert_bytes_keys_to_strings(item) for item in obj]
    else:
        return obj


def serialize(**kwargs) -> bytes:
    """Serializes a Python dict-like object for ZeroMQ transmission.

    Preserves named arguments as key-value pairs.

    Args:
        kwargs (dict): Inputs to serialize using a custom encoding hook for NumPy arrays.

    Returns:
        bytes: Serialized binary data safe to transmit.
    """
    return msgpack.packb(o=kwargs, default=encode_ndarray)  # type: ignore


def deserialize(msg: bytes) -> dict:
    """Deserializes the received message to construct the original Python object.

    Args:
        msg (bytes): Raw binary data containing the original Python object.

    Returns:
        dict: Python object with the key-value pairs preserved and any NumPy arrays reconstructed.
    """
    raw_dict = msgpack.unpackb(msg, object_hook=decode_ndarray)
    return convert_bytes_keys_to_strings(raw_dict)  # type: ignore



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from abc import ABC, abstractmethod
from collections import OrderedDict

from hermes.base.stream import Stream


class StorageInterface(ABC):
    """Interface for the AsyncIO Storage component."""

    @abstractmethod
    def _set_state(self, state) -> None:
        """User defined logic for FSM state transition.

        Args:
            state (Any): New state to transition to.
        """
        pass

    @abstractmethod
    def _initialize(self, streams: OrderedDict[str, Stream]) -> None:
        """Initializes files and indices for write pointer tracking.

        Args:
            streams (OrderedDict[str, Stream]): Reference to the Stream objects to flush to disk.
        """
        pass

    @abstractmethod
    async def _log_data(self) -> None:
        """Main AsyncIO loop for Storage to write files concurrentlyto disk."""
        pass

    @abstractmethod
    def _is_to_stream(self) -> bool:
        """Check if any streams were configured to stream.

        Returns:
            bool: Whether there are any streams configured to stream data.
        """
        pass

    @abstractmethod
    def _is_to_dump(self) -> bool:
        """Check if any streams were configured to record.

        Returns:
            bool: Whether there are any streams configured to dump record data.
        """
        pass

    @abstractmethod
    def _start_stream_logging(self) -> None:
        """Set up AV/HDF5 file writers for stream-logging, if desired."""
        pass

    @abstractmethod
    def _stop_stream_logging(self) -> None:
        """Trigger termination and flushing of accumulated streamed data.

        Will wait for the thread to finish before returning.
        """
        pass

    @abstractmethod
    def _start_dump_logging(self) -> None:
        """Initialize passive recording until terminated to dump data once."""
        pass

    @abstractmethod
    def _wait_till_flush(self) -> None:
        """Sleep until the Storage is triggered to terminate after flushing."""
        pass

    @abstractmethod
    def _release_thread_pool(self) -> None:
        """Trigger release of AsyncIO resources used for writing files."""
        pass



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from multiprocessing import Queue

from hermes.base.nodes.node import Node
from hermes.base.nodes.node_interface import NodeInterface
from hermes.utils.di_utils import search_module_class


def launch_node(spec: dict, input_queue: "Queue[tuple[float, str]]"):
    """Launches callable `Node` objects using the user-provided specification.

    Args:
        spec (dict): Specification containing at least package and `Node` names, and constructor arguments specific to that `Node`.
        input_queue (Queue[tuple[float, str]]): Multiprocessing queue to fan-in user keyboard inputs if the `Node` is interested to receive any.
    """
    module_name: str = spec["package"]
    class_name: str = spec["class"]
    class_args: dict = spec["settings"]
    node_class: type[NodeInterface] = search_module_class(module_name, class_name)
    node: Node = node_class(**class_args, input_queue=input_queue)
    node()



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

import time
import random

from hermes.utils.time_utils import get_time
from hermes.utils.zmq_utils import PORT_BACKEND, PORT_KILL, PORT_SYNC_HOST
from hermes.utils.types import LoggingSpec

from hermes.dummy.stream import DummyStream
from hermes.base.nodes.producer import Producer


class DummyProducer(Producer):
    """A Node showcasing the Producer behavior, generating new data relayed to the Broker for consumers."""

    @classmethod
    def _log_source_tag(cls) -> str:
        return "dummy-producer"

    def __init__(
        self,
        host_ip: str,
        logging_spec: LoggingSpec,
        sampling_rate_hz: int = 1,
        payload_num_bytes: int = 100,
        port_pub: str = PORT_BACKEND,
        port_sync: str = PORT_SYNC_HOST,
        port_killsig: str = PORT_KILL,
        transmit_delay_sample_period_s: float = float("nan"),
        **_,
    ):
        """Constructor of the DummyProducer Node.

        Args:
            host_ip (str): IP address of the local master Broker.
            logging_spec (LoggingSpec): Specification of what and how to store.
            sampling_rate_hz (int, optional): Expected sample rate of the device. Defaults to `1`.
            payload_num_bytes (int, optional): Size of the messages in bytes to generate. Defaults to `100`.
            port_pub (str, optional): Local port to publish to for local master Broker to relay. Defaults to `PORT_BACKEND`.
            port_sync (str, optional): Local port to listen to for local master Broker's startup coordination. Defaults to `PORT_SYNC_HOST`.
            port_killsig (str, optional): Local port to listen to for local master Broker's termination signal. Defaults to `PORT_KILL`.
            transmit_delay_sample_period_s (float, optional): Duration of the period over which to estimate propagation delay of measurements from the corresponding device. Defaults to `float('nan')`.
        """

        sampling_rate_hz = (
            sampling_rate_hz
            if isinstance(sampling_rate_hz, (int, float))
            else int(sampling_rate_hz)
        )
        payload_num_bytes = (
            payload_num_bytes
            if isinstance(payload_num_bytes, (int))
            else int(payload_num_bytes)
        )

        self._period = 1 / sampling_rate_hz
        self._payload_num_bytes = payload_num_bytes
        self._sequence = 0
        self._data = random.randbytes(self._payload_num_bytes)
        self._tag: str = "%s.data" % self._log_source_tag()
        self._next_period: float

        stream_out_spec = {
            "sampling_rate_hz": sampling_rate_hz,
            "payload_num_bytes": payload_num_bytes,
        }

        super().__init__(
            host_ip=host_ip,
            stream_out_spec=stream_out_spec,
            logging_spec=logging_spec,
            port_pub=port_pub,
            port_sync=port_sync,
            port_killsig=port_killsig,
            transmit_delay_sample_period_s=transmit_delay_sample_period_s,
        )

    @classmethod
    def create_stream(cls, stream_spec: dict) -> DummyStream:
        return DummyStream(**stream_spec)

    def _ping_device(self) -> None:
        return None

    def _connect(self) -> bool:
        return True

    def _keep_samples(self) -> None:
        self._next_period = get_time() + self._period

    def _process_data(self) -> None:
        if self._is_continue_capture:
            process_time_s = get_time()
            time_to_wait = self._next_period - process_time_s

            if time_to_wait > 0:
                time.sleep(time_to_wait * 0.9)
                while (process_time_s := get_time()) < self._next_period:
                    pass

            self._publish(
                self._tag,
                process_time_s=process_time_s,
                data={
                    "sensor-emulator": {"data": self._data, "sequence": self._sequence}
                },
            )
            self._sequence += 1
            self._next_period += self._period
        else:
            self._send_end_packet()

    def _stop_new_data(self):
        pass

    def _cleanup(self) -> None:
        super()._cleanup()



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from abc import abstractmethod
from collections import OrderedDict
import threading
import zmq
import math

from hermes.utils.msgpack_utils import serialize
from hermes.utils.zmq_utils import (
    CMD_END,
    CMD_EXIT,
    DNS_LOCALHOST,
    PORT_BACKEND,
    PORT_KILL,
    PORT_SYNC_HOST,
)
from hermes.utils.types import LoggingSpec

from hermes.base.stream import Stream
from hermes.base.storage.storage import Storage
from hermes.base.delay_estimator import DelayEstimator
from hermes.base.nodes.node import Node
from hermes.base.nodes.producer_interface import ProducerInterface


class Producer(ProducerInterface, Node):
    """An abstract class wrapping an interface with a particular device into a Producer Node."""

    def __init__(
        self,
        host_ip: str,
        stream_out_spec: dict,
        logging_spec: LoggingSpec,
        sampling_rate_hz: float = float("nan"),
        port_pub: str = PORT_BACKEND,
        port_sync: str = PORT_SYNC_HOST,
        port_killsig: str = PORT_KILL,
        transmit_delay_sample_period_s: float = float("nan"),
    ) -> None:
        """Constructor of the Producer parent class.

        Args:
            host_ip (str): IP address of the local master Broker.
            stream_out_spec (dict): Mapping of corresponding Stream object parameters to user-defined configuration values.
            logging_spec (LoggingSpec): Specification of what and how to store.
            sampling_rate_hz (float, optional): Expected sample rate of the device. Defaults to `float('nan')`.
            port_pub (str, optional): Local port to publish to for local master Broker to relay. Defaults to `PORT_BACKEND`.
            port_sync (str, optional): Local port to listen to for local master Broker's startup coordination. Defaults to `PORT_SYNC_HOST`.
            port_killsig (str, optional): Local port to listen to for local master Broker's termination signal. Defaults to `PORT_KILL`.
            transmit_delay_sample_period_s (float, optional): Duration of the period over which to estimate propagation delay of measurements from the corresponding device. Defaults to `float('nan')`.
        """
        super().__init__(
            host_ip=host_ip,
            port_sync=port_sync,
            port_killsig=port_killsig,
            ref_time=logging_spec.ref_time_s,
        )
        self._sampling_rate_hz = sampling_rate_hz
        self._sampling_period = 1 / sampling_rate_hz
        self._port_pub = port_pub
        self._is_continue_capture = True
        self._transmit_delay_sample_period_s = transmit_delay_sample_period_s
        self._publish_fn = lambda tag, **kwargs: None

        # Data structure for keeping track of data.
        self._stream: Stream = self.create_stream(stream_out_spec)

        # Create the data storing object.
        self._storage = Storage(self._log_source_tag(), logging_spec)

        # Launch datalogging thread with reference to the Stream object.
        self._storage_thread = threading.Thread(
            target=self._storage,
            args=(OrderedDict([(self._log_source_tag(), self._stream)]),),
        )
        self._storage_thread.start()

        # Conditional creation of the transmission delay estimate thread.
        if not math.isnan(self._transmit_delay_sample_period_s):
            self._delay_estimator = DelayEstimator(self._transmit_delay_sample_period_s)
            self._delay_thread = threading.Thread(
                target=self._delay_estimator,
                kwargs={
                    "ping_fn": self._ping_device,
                    "publish_fn": lambda time_s, delay_s: self._publish(
                        tag="%s.connection" % self._log_source_tag(),
                        time_s=time_s,
                        data={
                            "%s-connection"
                            % self._log_source_tag(): {"transmission_delay": delay_s}
                        },
                    ),
                },
            )
            self._delay_thread.start()

    def _publish(self, tag: str, **kwargs) -> None:
        """Common method to save and publish the captured sample.

        Best to deal with data structure (threading primitives) AFTER handing off packet to ZeroMQ.
        That way network thread can already start processing the packet.

        Args:
            tag (str): Uniquely identifying key for the modality to label data for message exchange.
        """
        self._publish_fn(tag, **kwargs)

    def _initialize(self):
        super()._initialize()
        # Socket to publish sensor data and log
        self._pub: zmq.SyncSocket = self._ctx.socket(zmq.PUB)
        self._pub.connect("tcp://%s:%s" % (DNS_LOCALHOST, self._port_pub))
        self._connect()

    def _activate_data_poller(self) -> None:
        self._poller.register(self._pub, zmq.POLLOUT)

    def _on_poll(self, poll_res):
        # Process custom event first, then Node generic (killsig).
        if self._pub in poll_res[0]:
            self._process_data()
        super()._on_poll(poll_res)

    def _on_sync_complete(self) -> None:
        self._publish_fn = self._store_and_broadcast
        self._keep_samples()

    def _store_and_broadcast(self, tag: str, process_time_s: float, **kwargs) -> None:
        """Place captured data into the corresponding Stream datastructure and transmit serialized ZeroMQ packets to subscribers.

        Args:
            tag (str): Uniquely identifying key for the modality to label data for message exchange.
        """
        msg = serialize(**kwargs)
        self._pub.send_multipart([tag.encode("utf-8"), msg])
        self._stream.append_data(process_time_s=process_time_s, **kwargs)

    def _trigger_stop(self):
        self._is_continue_capture = False
        self._stop_new_data()

    def _send_end_packet(self) -> None:
        """Send 'END' empty packet and label Node as done to safely finish and exit the process and its threads."""
        self._pub.send_multipart(
            [
                ("%s.data" % self._log_source_tag()).encode("utf-8"),
                CMD_END.encode("utf-8"),
            ]
        )
        self._is_done = True

    @abstractmethod
    def _cleanup(self) -> None:
        # Indicate to Storage to wrap up and exit.
        self._storage.cleanup()
        if not math.isnan(self._transmit_delay_sample_period_s):
            self._delay_estimator.cleanup()
        # Before closing the PUB socket, wait for the 'BYE' signal from the Broker.
        self._sync.send_multipart(
            [self._log_source_tag().encode("utf-8"), CMD_EXIT.encode("utf-8")]
        )
        host, cmd = (
            self._sync.recv_multipart()
        )  # no need to read contents of the message.
        print(
            "%s received %s from %s."
            % (self._log_source_tag(), cmd.decode("utf-8"), host.decode("utf-8")),
            flush=True,
        )
        self._pub.close()
        # Join on the logging background thread last, so that all things can finish in parallel.
        self._storage_thread.join()
        if not math.isnan(self._transmit_delay_sample_period_s):
            self._delay_thread.join()
        super()._cleanup()



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

import random
import string
from hermes.utils.time_utils import get_time
from hermes.utils.zmq_utils import (
    PORT_BACKEND,
    PORT_FRONTEND,
    PORT_KILL,
    PORT_SYNC_HOST,
)
from hermes.utils.types import LoggingSpec

from hermes.dummy.stream import DummyPipeStream
from hermes.base.nodes.pipeline import Pipeline


class DummyPipeline(Pipeline):
    """A Node showcasing the Pipeline behavior, consuming external data and generating new data relayed back to the Broker."""

    @classmethod
    def _log_source_tag(cls) -> str:
        return "dummy-pipeline"

    def __init__(
        self,
        host_ip: str,
        stream_out_spec: dict,
        stream_in_specs: list[dict],
        logging_spec: LoggingSpec,
        is_async_generate: bool = False,
        port_pub: str = PORT_BACKEND,
        port_sub: str = PORT_FRONTEND,
        port_sync: str = PORT_SYNC_HOST,
        port_killsig: str = PORT_KILL,
        **_,
    ):
        """Constructor of the DummyPipeline Node.

        Args:
            host_ip (str): IP address of the local master Broker.
            stream_out_spec (dict): Mapping of corresponding Stream object parameters to user-defined configuration values.
            stream_in_specs (list[dict]): List of mappings of user-configured incoming modalities.
            logging_spec (LoggingSpec): Specification of what and how to store.
            is_async_generate (bool, optional): Whether the Pipeline produces data asynchronously, in parallel to what is fed into it. Defaults to `False`.
            port_pub (str, optional): Local port to publish to for local master Broker to relay. Defaults to `PORT_BACKEND`.
            port_sub (str, optional): Local port to subscribe to for incoming relayed data from the local master Broker. Defaults to `PORT_FRONTEND`.
            port_sync (str, optional): Local port to listen to for local master Broker's startup coordination. Defaults to `PORT_SYNC_HOST`.
            port_killsig (str, optional): Local port to listen to for local master Broker's termination signal. Defaults to `PORT_KILL`.
        """
        self._is_continue_generate = True
        self._is_keep_samples = False
        self._sequence = 0
        self._period = 1 / stream_out_spec["sampling_rate_hz"]
        self._next_period: float

        super().__init__(
            host_ip=host_ip,
            stream_out_spec=stream_out_spec,
            stream_in_specs=stream_in_specs,
            logging_spec=logging_spec,
            is_async_generate=is_async_generate,
            port_pub=port_pub,
            port_sub=port_sub,
            port_sync=port_sync,
            port_killsig=port_killsig,
        )

    @classmethod
    def create_stream(cls, stream_spec: dict) -> DummyPipeStream:
        return DummyPipeStream(**stream_spec)

    def _keep_samples(self) -> None:
        self._is_keep_samples = True
        self._next_period = get_time() + self._period

    def _process_data(self, topic: str, msg: dict) -> None:
        process_time_s: float = get_time()
        tag: str = "%s.data" % self._log_source_tag()
        data = msg["data"]["sensor-emulator"]
        data["flag"] = 1
        self._publish(tag, process_time_s=process_time_s, data={"sensor-emulator-processed": data})

    def _generate_data(self) -> None:
        if self._is_keep_samples and self._is_continue_generate:
            process_time_s = get_time()
            if self._next_period <= process_time_s:
                tag: str = "%s.data" % self._log_source_tag()
                data = {
                    "data": ''.join([random.choice(string.printable) for _ in range(random.randint(1, 100))]).encode("ascii"),
                    "sequence": self._sequence,
                }
                self._publish(tag, process_time_s=process_time_s, data={"sensor-emulator-internal": data})
                self._sequence += 1
                self._next_period += self._period
        elif self._is_keep_samples and not self._is_continue_generate:
            self._notify_no_more_data_out()

    def _stop_new_data(self):
        self._is_continue_generate = False

    def _cleanup(self) -> None:
        super()._cleanup()



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from hermes.base.stream import Stream


class DummyStream(Stream):
    """A Stream structure to store Dummy modality data."""

    def __init__(
        self, sampling_rate_hz: int = 1, payload_num_bytes: int = 100, **_
    ) -> None:
        """Constructor of the DummyStream datastructure.

        Args:
            sampling_rate_hz (int, optional): Duration of the period over which new data becomes available. Defaults to `1`.
            payload_num_bytes (int, optional): Size of the messages to send. Defaults to `100`.
        """
        super().__init__()

        self._device_name = "sensor-emulator"

        self.add_stream(
            device_name=self._device_name,
            stream_name="sequence",
            data_type="uint32",
            sample_size=[1],
            sampling_rate_hz=int(sampling_rate_hz),
            is_measure_rate_hz=False,
        )
        self.add_stream(
            device_name=self._device_name,
            stream_name="data",
            data_type=f"S{payload_num_bytes}",
            sample_size=[1],
            sampling_rate_hz=int(sampling_rate_hz),
            is_measure_rate_hz=True,
        )

    def get_fps(self) -> dict[str, float | None]:
        return {self._device_name: super()._get_fps(self._device_name, "data")}


class DummyPipeStream(Stream):
    """A Stream structure to store Dummy Pipeline modality data."""

    def __init__(
        self, sampling_rate_hz: int = 1, incoming_payload_num_bytes: int = 100, **_,
    ) -> None:
        """Constructor of the DummyStream datastructure.

        Args:
            sampling_rate_hz (int, optional): Number of times per second, monotonically spaced, that new data becomes available. Defaults to `1`.
        """
        super().__init__()

        self.add_stream(
            device_name="sensor-emulator-processed",
            stream_name="sequence",
            data_type="uint32",
            sample_size=[1],
        )
        self.add_stream(
            device_name="sensor-emulator-processed",
            stream_name="data",
            data_type=f"S{incoming_payload_num_bytes}",
            sample_size=[1],
        )
        self.add_stream(
            device_name="sensor-emulator-processed",
            stream_name="flag",
            data_type="uint8",
            sample_size=[1],
        )

        self.add_stream(
            device_name="sensor-emulator-internal",
            stream_name="sequence",
            data_type="uint32",
            sample_size=[1],
            sampling_rate_hz=int(sampling_rate_hz),
        )
        self.add_stream(
            device_name="sensor-emulator-internal",
            stream_name="data",
            data_type="T",
            sample_size=[1],
            sampling_rate_hz=int(sampling_rate_hz),
            is_measure_rate_hz=True,
        )

    def get_fps(self) -> dict[str, float | None]:
        return {
            "sensor-emulator-processed": super()._get_fps("sensor-emulator-processed", "data"),
            "sensor-emulator-internal": super()._get_fps("sensor-emulator-internal", "data"),
        }



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from abc import ABC, abstractmethod
import zmq


class NodeInterface(ABC):
    """Interface for the Node components."""

    @classmethod
    @abstractmethod
    def _log_source_tag(cls) -> str:
        """Read-only property uniquely identifying the Node.

        Returns:
            str: Unique key identifying the Node in the data exchange.
        """
        pass

    @property
    @abstractmethod
    def _is_done(self) -> bool:
        """Read-only property identifying if the Node completed operation.

        Returns:
            bool: Whether the Node completed its function.
        """
        pass

    @abstractmethod
    def _set_state(self, state) -> None:
        """User-defined logic for FSM state transition.

        Args:
            state (Any): New state to transition to.
        """
        pass

    @abstractmethod
    def _initialize(self) -> None:
        """Node-specific initialization procedure.

        Pre-run setup of the backend specific to the Node implementaiton.
        Generic setup should be run first.
        """
        pass

    @abstractmethod
    def _get_sync_socket(self) -> zmq.SyncSocket:
        """Get reference to the socket used for synchronization of the Node to its master Broker.

        Returns:
            zmq.SyncSocket: ZeroMQ socket of the Node connected to the local master Broker.
        """
        pass

    @abstractmethod
    def _activate_kill_poller(self) -> None:
        """Start listening for the KILL signal on the special PUB/SUB socket that coordinates program termination."""
        pass

    @abstractmethod
    def _activate_data_poller(self) -> None:
        """Start listening for new data from other Nodes."""
        pass

    @abstractmethod
    def _deactivate_kill_poller(self) -> None:
        """Stop listening for the KILL signal."""
        pass

    @abstractmethod
    def _send_kill_to_broker(self) -> None:
        """Send a slave KILL signal to the local Broker in case program termination by the slave Node is recorded."""
        pass

    @abstractmethod
    def _poll(self) -> tuple[list[zmq.SyncSocket], list[int]]:
        """Block for new ZeroMQ data to collect at the Poller.

        Listens for events when new data is received from or when new data can be written to sockets,
        based on the active Poller settings of the Node implementation.

        Returns:
            tuple[list[zmq.SyncSocket], list[int]]: Result of listening on the sockets registered by the Poller.
        """
        pass

    @abstractmethod
    def _on_poll(self, poll_res: tuple[list[zmq.SyncSocket], list[int]]) -> None:
        """Callback to perform some logic everytime some data transactions are received by the Poller.

        Generic entry-point for all types of Nodes, based on their active Poller settings.
        NOTE: if Node in JoinState, kill socket is no longer in the Poller and only higher-level logic is triggered.

        Args:
            poll_res (tuple[list[zmq.SyncSocket], list[int]]): Reference to the complete captured result of listening by the Poller.
        """
        pass

    @abstractmethod
    def _trigger_stop(self) -> None:
        """Trigger to the Node's internal procedures and background threads to gracefully wrap-up.

        Producer: stops sampling data, continue sending already captured until none is left, with last message labeled 'END'.
        Consumer: continues listening to data until each of subscribed Producers sent the last message.
        Pipeline: continues listening to data to produce results until each data sources sent the last message, and then labels the last message with 'END'.
        """
        pass

    @abstractmethod
    def _on_sync_complete(self) -> None:
        """Callback to perform some logic after synchronization of Nodes is completed and indicated by the Broker."""
        pass



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from multiprocessing import Process, Queue
from multiprocessing.synchronize import Event as EventClass
from typing import Callable
import zmq

from hermes.utils.node_utils import launch_node
from hermes.utils.time_utils import get_time
from hermes.utils.types import ZMQResult
from hermes.utils.zmq_utils import (
    IP_LOOPBACK,
    PORT_BACKEND,
    PORT_FRONTEND,
    PORT_KILL,
    PORT_KILL_BTN,
    PORT_SYNC_HOST,
    PORT_SYNC_REMOTE,
    TOPIC_KILL,
)

from hermes.base.broker.broker_interface import BrokerInterface
from hermes.base.broker.broker_states import AbstractBrokerState, InitState


class Broker(BrokerInterface):
    """Manager of the lifecycle of all connected local Nodes and data broker.

    Facilitates high-performance message exchange using ZeroMQ zero-copy
    communication across distributed sensing and computing hosts.
    Passes data over sockets to external Brokers or over shared memory
    for local Nodes.

    Hosts control logic of interactive proxy/server.
    Will launch/destroy/connect to Nodes on creation and ad-hoc.
    Will use a separate process for each streamer and consumer.
    Each Node connects only to its local Broker,
        which then exposes its data to outside LAN subscribers.

    Uses fixed ports for communication under `zmq_utils.py`.
    Use of other ports is discouraged.

    Macro-defined ports are preferred for consistency in a
    distributed, multi-host setup.

    Uses hierarchical coordination of distributed host synchronization/setup.
    Each Broker first starts up and sync local Nodes.
    Then syncs with all expected remote hosts, until all are ready.
    After all are ready, master Broker communicates trigger signal to start streaming.
    """

    @classmethod
    def _log_source_tag(cls) -> str:
        return "manager"

    def __init__(
        self,
        host_ip: str,
        node_specs: list[dict],
        is_ready_event: EventClass,
        is_quit_event: EventClass,
        is_done_event: EventClass,
        is_master_broker: bool = False,
        port_backend: str = PORT_BACKEND,
        port_frontend: str = PORT_FRONTEND,
        port_sync_host: str = PORT_SYNC_HOST,
        port_sync_remote: str = PORT_SYNC_REMOTE,
        port_killsig: str = PORT_KILL,
    ) -> None:
        """Constructor of the Broker component responsible for the lifecycle of all local Nodes and for message exchange across them and distributed hosts.

        Args:
            host_ip (str): Public LAN IP address of this host.
            node_specs (list[dict]): List of to-be-created Node specification dictionaries.
            is_ready_event (Event): Synchronization primitive to indicate to external readers that Broker is ready.
            is_quit_event (Event): Synchronization primitive to externally trigger closure of Broker.
            is_done_event (Event): Synchronization primitive to indicate to external readers completion of Broker.
            is_master_broker (bool, optional): Whether this Broker is the master in the distributed host setup. Defaults to `False`.
            port_backend (str, optional): XSUB port of the Broker. Defaults to `PORT_BACKEND`.
            port_frontend (str, optional): XPUB port of the Broker. Defaults to `PORT_FRONTEND`.
            port_sync_host (str, optional): Port for SYNC socket to coordinate startup of local Nodes. Defaults to `PORT_SYNC_HOST`.
            port_sync_remote (str, optional): Port for SYNC socket to coordinate startup across remote hosts. Defaults to `PORT_SYNC_REMOTE`.
            port_killsig (str, optional): Port of the KILL signal this Broker announces from. Defaults to `PORT_KILL`.
        """
        self._host_ip = host_ip
        self._is_master_broker = is_master_broker
        self._port_backend = port_backend
        self._port_frontend = port_frontend
        self._port_sync_host = port_sync_host
        self._port_sync_remote = port_sync_remote
        self._port_killsig = port_killsig
        self._node_specs = node_specs

        self._is_ready_event = is_ready_event
        self._is_quit_event = is_quit_event
        self._is_done_event = is_done_event

        self._remote_pub_brokers: list[str] = []
        self._remote_sub_brokers: list[str] = []
        self._brokered_nodes: set[str] = set()

        self._processes: list[Process]
        self._queues: list[Queue[tuple[float, str]]] = [Queue() for _ in node_specs]

        # FSM for the broker
        self._state = InitState(self)

        # Pass exactly one ZeroMQ context instance throughout the program
        self._ctx: zmq.Context = zmq.Context()

        # Exposes a known address and port to locally connected sensors to connect to.
        local_backend: zmq.SyncSocket = self._ctx.socket(zmq.XSUB)
        local_backend.bind("tcp://%s:%s" % (IP_LOOPBACK, self._port_backend))
        self._backends: list[zmq.SyncSocket] = [local_backend]

        # Exposes a known address and port to broker data to local workers.
        local_frontend: zmq.SyncSocket = self._ctx.socket(zmq.XPUB)
        local_frontend.bind("tcp://%s:%s" % (IP_LOOPBACK, self._port_frontend))
        self._frontends: list[zmq.SyncSocket] = [local_frontend]

        # Listener endpoint to receive signals of streamers' readiness
        self._sync_host: zmq.SyncSocket = self._ctx.socket(zmq.ROUTER)
        self._sync_host.bind("tcp://%s:%s" % (self._host_ip, self._port_sync_host))

        # Socket to connect to remote Brokers
        self._sync_remote: zmq.SyncSocket = self._ctx.socket(zmq.ROUTER)
        self._sync_remote.setsockopt_string(
            zmq.IDENTITY, "%s:%s" % (self._host_ip, self._port_sync_remote)
        )
        self._sync_remote.bind("tcp://%s:%s" % (self._host_ip, self._port_sync_remote))

        # Termination control socket to command publishers and subscribers to finish and exit.
        killsig_pub: zmq.SyncSocket = self._ctx.socket(zmq.PUB)
        killsig_pub.bind("tcp://*:%s" % (self._port_killsig))
        self._killsigs: list[zmq.SyncSocket] = [killsig_pub]

        # Socket to listen to kill command from the GUI.
        self._gui_btn_kill: zmq.SyncSocket = self._ctx.socket(zmq.REP)
        self._gui_btn_kill.bind("tcp://*:%s" % PORT_KILL_BTN)

        # Poll object to listen to sockets without blocking
        self._poller: zmq.Poller = zmq.Poller()

    def expose_to_remote_broker(self, addr: list[str]) -> None:
        """Exposes a known address and port to remote networked subscribers if configured.

        Args:
            addr (list[str]): List of IP addresses of remote hosts (other Brokers).
        """
        frontend_remote: zmq.SyncSocket = self._ctx.socket(zmq.XPUB)
        frontend_remote.bind("tcp://%s:%s" % (self._host_ip, self._port_frontend))
        self._remote_sub_brokers.extend(addr)
        self._frontends.append(frontend_remote)

    def connect_to_remote_broker(
        self, addr: str, port_pub: str = PORT_FRONTEND
    ) -> None:
        """Connects to a known address and port of external LAN data broker.

        Args:
            addr (str): Remote host IP to connect to as a listener.
            port_pub (str, optional): Port number on which remote host publishes local Nodes' data. Defaults to `PORT_FRONTEND`.
        """
        backend_remote: zmq.SyncSocket = self._ctx.socket(zmq.XSUB)
        backend_remote.connect("tcp://%s:%s" % (addr, port_pub))
        self._remote_pub_brokers.append(addr)
        self._backends.append(backend_remote)

    def subscribe_to_killsig(self, addr: str, port_killsig: str = PORT_KILL) -> None:
        """Subscribes to external kill signal of another host as master.

        Args:
            addr (str): IP address of the master Broker in a distributed setting.
            port_killsig (str, optional): Port of the remote Broker to listen to for the termination signal. Defaults to `PORT_KILL`.
        """
        killsig_sub: zmq.SyncSocket = self._ctx.socket(zmq.SUB)
        killsig_sub.connect("tcp://%s:%s" % (addr, port_killsig))
        killsig_sub.subscribe(TOPIC_KILL)
        self._poller.register(killsig_sub, zmq.POLLIN)
        self._killsigs.append(killsig_sub)

    #####################
    ###### RUNNING ######
    #####################
    def __call__(self, duration_s: float | None = None) -> None:
        """The main FSM loop of the Broker.

        Runs continuously until the user ends the experiment or after the specified duration.
        The duration start to count only after all Nodes established communication and synced.

        Args:
            duration_s (float | None, optional): Duration of data capturing/streaming. Defaults to `None`.
        """
        self._duration_s = duration_s
        while self._state.is_continue() and not self._is_quit_event.is_set():
            self._state.run()
        if self._is_quit_event.is_set():
            print(
                "Keyboard exit signalled. Safely closing and saving, have some patience...",
                flush=True,
            )
        self._state.kill()
        # Continue with the FSM until it gracefully wraps up.
        while self._state.is_continue():
            self._state.run()
        self._stop()
        print("Experiment ended, thank you for using our system <3", flush=True)
        
    def _fanout_user_input(self, user_input: tuple[float, str]) -> None:
        """Forward user keyboard input from the main thread of the parent process, to all the subprocesses.

        Args:
            user_input (tuple[float, str]): Keyboard user input from the `input()` call.
        """
        for q in self._queues:
            q.put(user_input)

    #############################
    ###### GETTERS/SETTERS ######
    #############################
    def _set_state(self, state: AbstractBrokerState) -> None:
        self._state = state
        self._state_start_time_s = get_time()

    def _set_broker_ready(self) -> None:
        self._is_ready_event.set()

    def _set_node_addresses(self, node_addresses: dict[str, bytes]) -> None:
        self._node_addresses = node_addresses

    def _get_node_addresses(self) -> dict[str, bytes]:
        return self._node_addresses

    def _set_remote_broker_addresses(self, remote_brokers: dict[str, bytes]) -> None:
        self._remote_brokers = remote_brokers

    def _get_remote_broker_addresses(self) -> dict[str, bytes]:
        return self._remote_brokers

    def _get_start_time(self) -> float:
        return self._state_start_time_s

    def _get_duration(self) -> float | None:
        return self._duration_s

    def _get_num_local_nodes(self) -> int:
        return len(self._processes)

    def _get_num_frontends(self) -> int:
        return len(self._frontends)

    def _get_num_backends(self) -> int:
        return len(self._backends)

    def _get_remote_pub_brokers(self) -> list[str]:
        return self._remote_pub_brokers

    def _get_remote_sub_brokers(self) -> list[str]:
        return self._remote_sub_brokers

    def _get_is_master_broker(self) -> bool:
        return self._is_master_broker

    def _get_brokered_nodes(self) -> set[str]:
        return self._brokered_nodes

    def _add_brokered_node(self, topic: str) -> None:
        self._brokered_nodes.add(topic)

    def _remove_brokered_node(self, topic: str) -> None:
        self._brokered_nodes.remove(topic)

    def _get_host_ip(self) -> str:
        return self._host_ip

    def _get_sync_host_socket(self) -> zmq.SyncSocket:
        return self._sync_host

    def _get_sync_remote_socket(self) -> zmq.SyncSocket:
        return self._sync_remote

    def _get_poller(self) -> zmq.Poller:
        return self._poller

    def _activate_pubsub_poller(self) -> None:
        for s in self._backends:
            self._poller.register(s, zmq.POLLIN)
        for s in self._frontends:
            self._poller.register(s, zmq.POLLIN)
        # Register KILL_BTN port REP socket with POLLIN event.
        self._poller.register(self._gui_btn_kill, zmq.POLLIN)

    def _deactivate_pubsub_poller(self) -> None:
        for s in self._backends:
            self._poller.unregister(s)
        for s in self._frontends:
            self._poller.unregister(s)

    def _start_local_nodes(self) -> None:
        self._processes: list[Process] = [
            Process(target=launch_node, args=(node_spec, input_queue))
            for node_spec, input_queue in zip(self._node_specs, self._queues)
        ]
        for p in self._processes:
            p.start()

    def _poll(self, timeout_ms: int) -> ZMQResult:
        return self._poller.poll(timeout=timeout_ms)

    def _broker_packets(
        self,
        poll_res: ZMQResult,
        on_data_received: Callable[[list[bytes]], None] = lambda _: None,
        on_subscription_changed: Callable[[list[bytes]], None] = lambda _: None,
    ) -> None:
        for recv_socket, _ in poll_res:
            # Forwards data packets from publishers to subscribers.
            if recv_socket in self._backends:
                msg = recv_socket.recv_multipart()
                on_data_received(msg)
                for send_socket in self._frontends:
                    send_socket.send_multipart(msg)
            # Forwards subscription packets from subscribers to publishers.
            if recv_socket in self._frontends:
                msg = recv_socket.recv_multipart()
                on_subscription_changed(msg)
                for send_socket in self._backends:
                    send_socket.send_multipart(msg)

    def _check_for_kill(self, poll_res: ZMQResult) -> bool:
        for sock, _ in poll_res:
            # Receives KILL from the GUI.
            if sock == self._gui_btn_kill:
                return True
            # Receives KILL signal from another broker.
            elif sock in self._killsigs:
                return True
        return False

    def _publish_kill(self) -> None:
        for kill_socket in self._killsigs[1:]:
            # Ignore any more KILL signals, enter the wrap-up routine.
            self._poller.unregister(kill_socket)
        # Ignore poll events from the GUI and the same socket if used by child processes to indicate keyboard interrupt.
        self._poller.unregister(self._gui_btn_kill)
        # Send kill signals to own locally connected devices.
        self._killsigs[0].send(TOPIC_KILL.encode("utf-8"))

    def _stop(self) -> None:
        """Gracefully exit after all local subprocesses terminate, and cleanup."""
        for p in self._processes:
            p.join()

        # Release all used local sockets.
        for s in self._backends:
            s.close()
        for s in self._frontends:
            s.close()
        for s in self._killsigs:
            s.close()
        self._sync_host.close()
        self._sync_remote.close()
        self._gui_btn_kill.close()

        # Destroy ZeroMQ context.
        self._ctx.term()
        self._is_done_event.set()



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

import time
import zmq

from hermes.utils.time_utils import get_time
from hermes.utils.types import ZMQResult
from hermes.utils.zmq_utils import (
    CMD_ACK,
    CMD_BYE,
    CMD_END,
    CMD_GO,
    CMD_HELLO,
    CMD_START_TIME,
    PORT_SYNC_REMOTE,
)

from hermes.base.broker.broker_interface import BrokerInterface
from hermes.base.state_interface import StateInterface


class AbstractBrokerState(StateInterface):
    """Abstract class for the Broker FSM.

    Can be externally triggered into the KILL state from any child State class.
    """

    def __init__(self, context: BrokerInterface):
        """Constructor of the AbstractBrokerState class.

        Args:
            context (BrokerInterface): Reference to the Broker object.
        """
        self._context = context

    def is_continue(self) -> bool:
        return True

    def kill(self) -> None:
        self._context._set_state(KillState(self._context))


class InitState(AbstractBrokerState):
    """Initialization state of the Broker to launch local Nodes.

    Activates broker poller sockets and goes in sync to wait for local Nodes to start up.
    """

    def run(self) -> None:
        self._context._activate_pubsub_poller()
        self._context._start_local_nodes()
        self._context._set_state(SyncNodeBarrierState(self._context))


class SyncNodeBarrierState(AbstractBrokerState):
    """Synchronization state of the Broker that waits until all local Nodes setup.

    Waits until all local Nodes signalled that they are initialized and ready to go.
    """

    def run(self) -> None:
        host_ip: str = self._context._get_host_ip()
        sync_host_socket: zmq.SyncSocket = self._context._get_sync_host_socket()
        num_left_to_sync: int = self._context._get_num_local_nodes()
        node_addresses = dict()
        while num_left_to_sync:
            address, _, node_name, cmd = sync_host_socket.recv_multipart()
            num_left_to_sync -= 1
            node_name = node_name.decode("utf-8")
            node_addresses[node_name] = address
            print(
                "%s connected to %s with %s message."
                % (node_name, host_ip, cmd.decode("utf-8")),
                flush=True,
            )
        self._context._set_node_addresses(node_addresses)
        self._context._set_state(SyncBrokerBarrierState(self._context))


class SyncBrokerBarrierState(AbstractBrokerState):
    """Synchronization state of the Broker that waits until all remote Brokers setup.

    Communicate to other Brokers that every local device is ready.
    """

    def __init__(self, context: BrokerInterface):
        super().__init__(context)
        self._host_ip: str = self._context._get_host_ip()
        self._sync_remote_socket: zmq.SyncSocket = (
            self._context._get_sync_remote_socket()
        )

        self._remote_sub_brokers = self._context._get_remote_sub_brokers()
        self._remote_pub_brokers = self._context._get_remote_pub_brokers()
        self._brokers = dict()

        self._brokers_left_to_acknowledge = set(self._remote_sub_brokers)
        self._brokers_left_to_checkin = set(self._remote_pub_brokers)
        for ip in self._remote_sub_brokers:
            self._sync_remote_socket.connect("tcp://%s:%s" % (ip, PORT_SYNC_REMOTE))
        # Register remote SYNC socket to receive requests from remote publishers,
        #   and responses from remote subscribers.
        self._poller = zmq.Poller()
        self._poller.register(self._sync_remote_socket, zmq.POLLIN)

    def run(self) -> None:
        # I am remote publishing Broker, I must notify subscribing Brokers that I am ready.
        for ip in self._remote_sub_brokers:
            self._sync_remote_socket.send_multipart(
                [
                    ("%s:%s" % (ip, PORT_SYNC_REMOTE)).encode("utf-8"),
                    b"",
                    self._host_ip.encode("utf-8"),
                    CMD_HELLO.encode("utf-8"),
                ]
            )

        # Check every 5 seconds if other Brokers completed their setup and responded back.
        # Could be that no other Brokers exist.
        poll_res: list[tuple[zmq.SyncSocket, zmq.PollEvent]]
        if poll_res := self._poller.poll(5000):  # type: ignore
            socket, _ = poll_res[0]
            address, _, broker_name, cmd = socket.recv_multipart()
            broker_name = broker_name.decode("utf-8")
            print(
                "%s sent %s to %s"
                % (broker_name, cmd.decode("utf-8"), self._host_ip),
                flush=True,
            )
            if broker_name in self._brokers_left_to_acknowledge:
                # Remote publisher received ACK from remote subscriber.
                self._brokers_left_to_acknowledge.remove(broker_name)
                self._brokers[broker_name] = address
            elif broker_name in self._brokers_left_to_checkin:
                self._brokers_left_to_checkin.remove(broker_name)
                self._brokers[broker_name] = address
                # Remote subscriber responds with ACK to remote publisher.
                self._sync_remote_socket.send_multipart(
                    [
                        address,
                        b"",
                        self._host_ip.encode("utf-8"),
                        CMD_ACK.encode("utf-8"),
                    ]
                )

        # Proceed to the next state to agree on the common time once all Brokers synchronized.
        if not self._brokers_left_to_acknowledge and not self._brokers_left_to_checkin:
            self._poller.unregister(self._sync_remote_socket)
            self._context._set_remote_broker_addresses(self._brokers)
            self._context._set_state(StartState(self._context))


class StartState(AbstractBrokerState):
    """Start state of the Broker that waits or initiates distributed launch.

    Trigger local Nodes to start logging when the agreed start time arrives.
    """

    def run(self) -> None:
        # If current Broker is not Master, wait for the SYNC signal with time when to start.
        host_ip: str = self._context._get_host_ip()
        sync_remote_socket: zmq.SyncSocket = self._context._get_sync_remote_socket()
        sync_host_socket: zmq.SyncSocket = self._context._get_sync_host_socket()
        nodes: dict[str, bytes] = self._context._get_node_addresses()
        brokers: dict[str, bytes] = self._context._get_remote_broker_addresses()

        # Master Broker selects start time as 5 seconds from now and distributes across Brokers.
        if self._context._get_is_master_broker():
            start_time_s: int = round(get_time()) + 5
            for address in brokers.values():
                sync_remote_socket.send_multipart(
                    [
                        address,
                        b"",
                        CMD_START_TIME.encode("utf-8"),
                        start_time_s.to_bytes(length=4, byteorder="big"),
                    ]
                )
        # Slave Brokers block on the reeceive socket, waiting for the time.
        else:
            address, _, cmd, start_time_bytes = sync_remote_socket.recv_multipart()
            start_time_s = int.from_bytes(start_time_bytes, byteorder="big")

        # Each Broker waits until that time comes to trigger start of logging, with 1ms precision.
        while (current_time_s := get_time()) < start_time_s:
            time.sleep(min(0.001, start_time_s - current_time_s))

        # Trigger local Nodes to start logging.
        for name, address in list(nodes.items()):
            sync_host_socket.send_multipart(
                [address, b"", host_ip.encode("utf-8"), CMD_GO.encode("utf-8")]
            )
            print("%s sending %s to %s" % (host_ip, CMD_GO, name), flush=True)

        self._context._set_state(RunningState(self._context))


class RunningState(AbstractBrokerState):
    """Stead-state capturing or streaming state of the Broker.

    Will run until the the experiment is stopped or after a fixed period, if provided.
    """

    def __init__(self, context: BrokerInterface):
        super().__init__(context)
        context._set_broker_ready()
        print("\n\n### %s ###\n\n" % (CMD_GO), flush=True)
        if (duration_s := self._context._get_duration()) is not None:
            self._is_continue_fn = lambda: get_time() < (
                self._context._get_start_time() + duration_s
            )
        else:
            self._is_continue_fn = lambda: True

    def run(self) -> None:
        poll_res: ZMQResult = self._context._poll(1000)
        self._context._broker_packets(
            poll_res, on_subscription_changed=self._on_subscription_added
        )
        if self._context._check_for_kill(poll_res):
            self.kill()

    def is_continue(self) -> bool:
        return self._is_continue_fn()

    def _on_subscription_added(self, msg: list[bytes]) -> None:
        """Update a list on the Broker that keeps track of which Nodes are being brokered for."""
        topic: str = msg[0].decode("utf-8").split("\x01")[1]
        self._context._add_brokered_node(topic=topic)


class KillState(AbstractBrokerState):
    """Received 1 of 3 possible KILL signals to terminate.

    Relay it to all Nodes and Brokers and wrap up gracefully.
        from the local Keyboard Interrupt;
        from the Master Broker;
        from the GUI;
    """

    def run(self) -> None:
        self._context._publish_kill()
        self._context._set_state(JoinNodeBarrierState(self._context))

    # Override default kill function behavior because we are already in the killing process
    def kill(self) -> None:
        pass


class JoinNodeBarrierState(AbstractBrokerState):
    """Waits until all local Nodes send final packets then quits itself.

    Wait for all processes (local and remote) to send the last messages before closing.
    Continue brokering packets until signalled by all publishers that there will be no more packets.
    Append a frame to the ZeroMQ message that indicates the last message from the sensor.
    """

    def __init__(self, context: BrokerInterface):
        super().__init__(context)
        self._host_ip = self._context._get_host_ip()
        self._nodes = self._context._get_node_addresses()
        self._nodes_waiting_to_exit: set[str] = set()
        self._nodes_expected_end_pub_packet: set[str] = (
            self._context._get_brokered_nodes()
        )
        self._sync_host_socket: zmq.SyncSocket = self._context._get_sync_host_socket()
        self._poller = self._context._get_poller()
        self._poller.register(self._sync_host_socket, zmq.POLLIN)

    def run(self) -> None:
        poll_res: ZMQResult = self._context._poll(1000)
        # Brokers packets and releases local Producer Nodes in a callback once it published the end packet.
        self._context._broker_packets(poll_res, on_data_received=self._on_is_end_packet)
        # Checks if poll event was triggered by a local Node initiating closing.
        self._check_host_sync_socket(poll_res)
        # Proceed to exiting once all local Nodes finished.
        if self._is_finished():
            self._poller.unregister(self._sync_host_socket)
            self._context._deactivate_pubsub_poller()
            self._context._set_state(JoinBrokerBarrierState(self._context))

    def _on_is_end_packet(self, msg: list[bytes]) -> None:
        """Callback to track brokering of last packets of local Producers and Pipelines.

        Will get trigerred at most once per Node because publishing Nodes send it only once.
        """
        #   Once the Broker registers arrival of 'END' packet from a local Producer/Pipeline,
        #     it will signal 'BYE' to it to allow it to exit.
        if CMD_END.encode("utf-8") in msg:
            # Check if the END packet came from the Broker's scope, (one of the Broker's local Nodes).
            #   Continue brokering packets if just proxing it (not Broker's local Nodes).
            topic = msg[0].decode().split(".")[0]
            if self._nodes_expected_end_pub_packet:
                self._nodes_expected_end_pub_packet.remove(topic)
                # Allow local Producer/Pipeline to exit.
                self._release_local_node(topic)

    def _check_host_sync_socket(self, poll_res: ZMQResult) -> None:
        """Check if a local Node sent a request on the SYNC socket to indicate its closure.

        Can be triggered by all local Nodes: Producer, Consumer, or Pipeline, sending 'EXIT?' request.
        """
        for sock, _ in poll_res:
            if sock == self._sync_host_socket:
                address, _, node_name, cmd = self._sync_host_socket.recv_multipart()
                topic = node_name.decode("utf-8")
                print(
                    "%s received %s from %s" % (self._host_ip, cmd, topic), flush=True
                )
                self._nodes_waiting_to_exit.add(topic)
                self._release_local_node(topic)

    def _release_local_node(self, topic: str) -> None:
        """Release the local Node from the list of active Nodes of the Broker."""
        if (
            topic in self._nodes_waiting_to_exit
            and topic not in self._nodes_expected_end_pub_packet
        ):
            self._sync_host_socket.send_multipart(
                [
                    self._nodes[topic],
                    b"",
                    self._host_ip.encode("utf-8"),
                    CMD_BYE.encode("utf-8"),
                ]
            )
            del self._nodes[topic]
            self._nodes_waiting_to_exit.remove(topic)

    def _is_finished(self) -> bool:
        """Convenience method indicating if any local Nodes remain active.

        Will wait until all local Nodes finish.

        Returns:
            bool: Whether all local Nodes gracefully terminated.
        """
        return not self._nodes

    def kill(self) -> None:
        pass


class JoinBrokerBarrierState(AbstractBrokerState):
    """Waits until all remote Brokers, if any, have their local Nodes gracefully terminated.

    Wait for all dependent remote Brokers to send the acknowledgement messages that they
    no longer dependend on this Broker's data.

    Continue brokering packets until signalled by all remote hosts that there will be no more packets.
    Use the remote SYNC socket to coordinate when every Broker can exit.
    """

    def __init__(self, context: BrokerInterface):
        super().__init__(context)
        self._host_ip = self._context._get_host_ip()
        self._brokers = self._context._get_remote_broker_addresses()
        self._sync_remote_socket: zmq.SyncSocket = (
            self._context._get_sync_remote_socket()
        )

        self._remote_sub_brokers = self._context._get_remote_sub_brokers()
        self._remote_pub_brokers = self._context._get_remote_pub_brokers()

        self._brokers_left_to_acknowledge = set(self._remote_sub_brokers)
        self._brokers_left_to_checkin = set(self._remote_pub_brokers)

        self._poller = zmq.Poller()
        self._poller.register(self._sync_remote_socket, zmq.POLLIN)

    def run(self):
        # Notify Brokers that listen to our data that we are done and ready to exit as soon as they received all last data from us.
        for ip in self._remote_sub_brokers:
            self._sync_remote_socket.send_multipart(
                [
                    ("%s:%s" % (ip, PORT_SYNC_REMOTE)).encode("utf-8"),
                    b"",
                    self._host_ip.encode("utf-8"),
                    CMD_BYE.encode("utf-8"),
                ]
            )

        # Check every 5 seconds if other Brokers completed their cleanup and responded back ready to exit.
        poll_res: list[tuple[zmq.SyncSocket, zmq.PollEvent]]
        if poll_res := self._poller.poll(5000):  # type: ignore
            socket, _ = poll_res[0]
            address, _, broker_name, cmd = socket.recv_multipart()
            broker_name = broker_name.decode("utf-8")
            print(
                "%s sent %s to %s." % (broker_name, cmd.decode("utf-8"), self._host_ip),
                flush=True,
            )

            if broker_name in self._brokers_left_to_acknowledge:
                # Remote subscriber responded with ACK to us.
                self._brokers_left_to_acknowledge.remove(broker_name)
                self._brokers.pop(broker_name)
            elif broker_name in self._brokers_left_to_checkin:
                # Remote publisher sent a BYE request, respond with an ACK.
                self._brokers_left_to_checkin.remove(broker_name)
                self._brokers.pop(broker_name)
                self._sync_remote_socket.send_multipart(
                    [
                        address,
                        b"",
                        self._host_ip.encode("utf-8"),
                        CMD_BYE.encode("utf-8"),
                    ]
                )

    def is_continue(self) -> bool:
        return not not self._brokers

    def kill(self) -> None:
        pass



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

import threading
from abc import abstractmethod
from collections import OrderedDict
import zmq

from hermes.utils.time_utils import get_time
from hermes.utils.msgpack_utils import deserialize
from hermes.utils.di_utils import search_module_class
from hermes.utils.zmq_utils import (
    CMD_END,
    CMD_EXIT,
    DNS_LOCALHOST,
    PORT_FRONTEND,
    PORT_KILL,
    PORT_SYNC_HOST,
)
from hermes.utils.types import LoggingSpec

from hermes.base.stream import Stream
from hermes.base.storage.storage import Storage
from hermes.base.nodes.node import Node
from hermes.base.nodes.consumer_interface import ConsumerInterface
from hermes.base.nodes.producer_interface import ProducerInterface
from hermes.base.nodes.pipeline_interface import PipelineInterface


class Consumer(ConsumerInterface, Node):
    """An abstract class to interface with a particular data consumer.

    Subscribes to the modalities specified in and parametrized by `stream_in_specs`.
    """

    def __init__(
        self,
        host_ip: str,
        stream_in_specs: list[dict],
        logging_spec: LoggingSpec,
        port_sub: str = PORT_FRONTEND,
        port_sync: str = PORT_SYNC_HOST,
        port_killsig: str = PORT_KILL,
    ) -> None:
        """Constructor of the Consumer parent class.

        Args:
            host_ip (str): IP address of the local master Broker.
            stream_in_specs (list[dict]): List of mappings of user-configured incoming modalities.
            logging_spec (LoggingSpec): Specification of what and how to store.
            port_sub (str, optional): Local port to subscribe to for incoming relayed data from the local master Broker. Defaults to `PORT_FRONTEND`.
            port_sync (str, optional): Local port to listen to for local master Broker's startup coordination. Defaults to `PORT_SYNC_HOST`.
            port_killsig (str, optional): Local port to listen to for local master Broker's termination signal. Defaults to `PORT_KILL`.
        """
        super().__init__(
            host_ip=host_ip,
            port_sync=port_sync,
            port_killsig=port_killsig,
            ref_time=logging_spec.ref_time_s,
        )
        self._port_sub = port_sub
        self._is_producer_ended: OrderedDict[str, bool] = OrderedDict()
        self._poll_data_fn = self._poll_data_packets

        # Instantiate all desired Streams that the Consumer will subscribe to.
        self._streams: OrderedDict[str, Stream] = OrderedDict()
        for stream_spec in stream_in_specs:
            module_name: str = stream_spec["package"]
            class_name: str = stream_spec["class"]
            specs: dict = stream_spec["settings"]
            # Create the stream datastructure.
            class_type: type[ProducerInterface] | type[PipelineInterface] = search_module_class(module_name, class_name)
            class_object: Stream = class_type.create_stream(specs)
            # Store the streamer object.
            self._streams.setdefault(class_type._log_source_tag(), class_object)
            self._is_producer_ended.setdefault(class_type._log_source_tag(), False)

        # Create the data storing object.
        self._storage = Storage(self._log_source_tag(), logging_spec)
        # Launch datalogging thread with reference to the Stream object.
        self._storage_thread = threading.Thread(
            target=self._storage, args=(self._streams,)
        )
        self._storage_thread.start()

    def _initialize(self):
        super()._initialize()
        # Socket to subscribe to Producers
        self._sub: zmq.SyncSocket = self._ctx.socket(zmq.SUB)
        self._sub.connect("tcp://%s:%s" % (DNS_LOCALHOST, self._port_sub))

        # Subscribe to topics for each mentioned local and remote Nodes
        for tag in self._streams.keys():
            self._sub.subscribe(tag)

    # Launch data receiving.
    def _activate_data_poller(self) -> None:
        self._poller.register(self._sub, zmq.POLLIN)

    # Process custom event first, then Node generic (killsig).
    def _on_poll(self, poll_res):
        if self._sub in poll_res[0]:
            self._poll_data_fn()
        super()._on_poll(poll_res)

    def _on_sync_complete(self) -> None:
        pass

    def _poll_data_packets(self) -> None:
        """Receive data packets in a steady state.

        Gets called every time one of the requestes modalities produced new data.
        In normal operation mode, all messages are 2-part.
        """
        topic, payload = self._sub.recv_multipart()
        receive_time = get_time()
        msg = deserialize(payload)
        topic_tree: list[str] = topic.decode("utf-8").split(".")
        self._streams[topic_tree[0]].append_data(process_time_s=receive_time, **msg)

    def _poll_ending_data_packets(self) -> None:
        """Receive data packets from producers and monitor for end-of-stream signal.

        When system triggered a safe exit, Pipeline gets a mix of normal 2-part messages
        and 3-part 'END' message from each Producer that safely exited.
        It's more efficient to dynamically switch the callback instead of checking every message.

        Processes packets on each modality until all data sources sent the 'END' packet.
        If triggered to stop and no more available data, sends empty 'END' packet and joins.
        """
        topic, payload = self._sub.recv_multipart()
        receive_time = get_time()
        # 'END' empty packet from a Producer.
        if CMD_END.encode("utf-8") in payload:
            topic_tree: list[str] = topic.decode("utf-8").split(".")
            self._is_producer_ended[topic_tree[0]] = True
            if all(list(self._is_producer_ended.values())):
                self._is_done = True
        # Regular data packets.
        else:
            msg = deserialize(payload)
            topic_tree: list[str] = topic.decode("utf-8").split(".")
            self._streams[topic_tree[0]].append_data(process_time_s=receive_time, **msg)

    def _trigger_stop(self):
        self._poll_data_fn = self._poll_ending_data_packets

    @abstractmethod
    def _cleanup(self):
        self._storage.cleanup()
        self._storage_thread.join()
        # Before closing the PUB socket, wait for the 'BYE' signal from the Broker.
        self._sync.send_multipart(
            [self._log_source_tag().encode("utf-8"), CMD_EXIT.encode("utf-8")]
        )
        host, cmd = (
            self._sync.recv_multipart()
        )  # no need to read contents of the message.
        print(
            "%s received %s from %s."
            % (self._log_source_tag(), cmd.decode("utf-8"), host.decode("utf-8")),
            flush=True,
        )
        self._sub.close()
        super()._cleanup()



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from abc import ABC, abstractmethod
from typing import Callable
import zmq

from hermes.utils.types import ZMQResult


class BrokerInterface(ABC):
    """Interface for the Broker component."""

    @classmethod
    @abstractmethod
    def _log_source_tag(cls) -> str:
        """Read-only class property identifying the component.

        Returns:
            str: Unique identifier.
        """
        pass

    @abstractmethod
    def _start_local_nodes(self) -> None:
        """Spawn specified locally hosted Nodes, each in a separate process."""
        pass

    @abstractmethod
    def _set_state(self, state) -> None:
        """User defined logic for FSM state transition.

        Args:
            state (Any): New state to transition to.
        """
        pass

    @abstractmethod
    def _set_broker_ready(self) -> None:
        """Hook for FSM to set Broker's internal state as ready."""
        pass

    @abstractmethod
    def _get_num_local_nodes(self) -> int:
        """Get the number of Nodes hosted and managed by this Broker.

        Returns:
            int: Number of locally attached Nodes.
        """
        pass

    @abstractmethod
    def _get_num_frontends(self) -> int:
        """Get the number of XPUB interfaces to remote hosts, listening to the Broker.

        Returns:
            int: Number of sockets.
        """
        pass

    @abstractmethod
    def _get_num_backends(self) -> int:
        """Get the number of XSUB interfaces to remote hosts, the Broker listens to.

        Returns:
            int: Number of sockets.
        """
        pass

    @abstractmethod
    def _get_remote_pub_brokers(self) -> list[str]:
        """Get the list of remote publishing Brokers' IPs.

        Returns:
            list[str]: List of IP addresses.
        """
        pass

    @abstractmethod
    def _get_remote_sub_brokers(self) -> list[str]:
        """Get the list of remote subscribing Brokers' IPs.

        Returns:
            list[str]: List of IP addresses.
        """
        pass

    @abstractmethod
    def _get_is_master_broker(self) -> bool:
        """Get the master status of this Broker in a distributed setup.

        Returns:
            bool: Whether this Broker is the master.
        """
        pass

    @abstractmethod
    def _get_brokered_nodes(self) -> set[str]:
        """Get the set of unique local Node identifiers that Broker manages.

        Returns:
            set[str]: Set of unique identifiers.
        """
        pass

    @abstractmethod
    def _add_brokered_node(self, topic: str) -> None:
        """Add a unique local Node identifier joining the exchange via the Broker.

        Noes uniquely self-identify by the data topic they produce.

        Args:
            topic (str): Unique identifier of the Node.
        """
        pass

    @abstractmethod
    def _remove_brokered_node(self, topic: str) -> None:
        """Remove the existing local Node identifier from the exchange via the Broker.

        Args:
            topic (str): Unique identifier of the Node.
        """
        pass

    @abstractmethod
    def _get_start_time(self) -> float:
        """Get the start time when the Broker set everything up and started streaming.

        Useful for measuring run time of the experiment,
        excluding the lengthy setup process.

        Returns:
            float: Time in seconds since epoch.
        """
        pass

    @abstractmethod
    def _get_duration(self) -> float | None:
        """Get the user-requested active duration of the capture/streaming.

        Returns:
            float | None: Time in seconds, if specified on Broker launch.
        """
        pass

    @abstractmethod
    def _get_sync_host_socket(self) -> zmq.SyncSocket:
        """Get the reference to the RCV socket for syncing local Nodes.

        Returns:
            zmq.SyncSocket: ZeroMQ socket used to communicate SYNC process.
        """
        pass

    @abstractmethod
    def _get_sync_remote_socket(self) -> zmq.SyncSocket:
        """Get the reference to the RCV socket for syncing remote Brokers.

        Returns:
            zmq.SyncSocket: ZeroMQ socket used to communicate SYNC process.
        """
        pass

    @abstractmethod
    def _set_node_addresses(self, node_addresses: dict[str, bytes]) -> None:
        """Bulk-set socket identifiers and unique Node identifiers for Broker's local Nodes.

        Args:
            node_addresses (dict[str, bytes]): Mapping of unique Node topics and their ZeroMQ socket identifiers.
        """
        pass

    @abstractmethod
    def _get_node_addresses(self) -> dict[str, bytes]:
        """Bulk-get socket identifiers and unique Node identifiers for Broker's local Nodes.

        Returns:
            dict[str, bytes]: Mapping of unique Node topics and their ZeroMQ socket identifiers.
        """
        pass

    @abstractmethod
    def _set_remote_broker_addresses(self, remote_brokers: dict[str, bytes]) -> None:
        """Bulk-set socket identifiers and IP addresses of remote Brokers for this Broker.

        Args:
            remote_brokers (dict[str, bytes]): Mapping of remote IPs and their ZeroMQ socket identifiers.
        """
        pass

    @abstractmethod
    def _get_remote_broker_addresses(self) -> dict[str, bytes]:
        """Bulk-get socket identifiers and IP addresses of remote Brokers.

        Returns:
            dict[str, bytes]: Mapping of remote IPs and their ZeroMQ socket identifiers.
        """
        pass

    @abstractmethod
    def _get_host_ip(self) -> str:
        """Get the Broker's host machine LAN IP address.

        Returns:
            str: Host's IP address.
        """
        pass

    @abstractmethod
    def _activate_pubsub_poller(self) -> None:
        """Register PUB-SUB sockets on both Broker interfaces for polling."""
        pass

    @abstractmethod
    def _deactivate_pubsub_poller(self) -> None:
        """Stop listening on the PUB or SUB interfaces for new data packets."""
        pass

    @abstractmethod
    def _get_poller(self) -> zmq.Poller:
        """Get the ZeroMQ Poller object responsible for socket management.

        Returns:
            zmq.Poller: ZeroMQ poller to (de)activate listening on an interface.
        """
        pass

    @abstractmethod
    def _poll(self, timeout_ms: int) -> ZMQResult:
        """Block until any new packets are available on PUB or SUB Broker interfaces.

        Args:
            timeout_ms (int): Polling timeout duration to re-evaluate check for manual CLI termination.

        Returns:
            ZMQResult: New ZeroMQ packets from PUB or SUB interfaces.
        """
        pass

    @abstractmethod
    def _broker_packets(
        self,
        poll_res: ZMQResult,
        on_data_received: Callable[[list[bytes]], None] = lambda _: None,
        on_subscription_changed: Callable[[list[bytes]], None] = lambda _: None,
    ) -> None:
        """Move packets between publishers and subscribers, local and remote.

        Args:
            poll_res (ZMQResult): New ZeroMQ packets from PUB or SUB interfaces.
            on_data_received (_type_, optional): Callback for data packets. Defaults to lambda_:None.
            on_subscription_changed (_type_, optional): Callback for subscription status packets. Defaults to lambda_:None.
        """
        pass

    @abstractmethod
    def _check_for_kill(self, poll_res: ZMQResult) -> bool:
        """Check if received packets contain a kill signal from a downstream Broker.

        Args:
            poll_res (ZMQResult): New ZeroMQ packets from PUB or SUB interfaces.

        Returns:
            bool: Whether a KILL message is contained in messages.
        """
        pass

    @abstractmethod
    def _publish_kill(self):
        """Send kill signals to upstream Brokers and local Nodes."""
        pass



############
#
# Copyright (c) 2022 MIT CSAIL and Joseph DelPreto
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# See https://action-net.csail.mit.edu for more usage information.
# Created 2021-2022 for the MIT ActionNet project by Joseph DelPreto [https://josephdelpreto.com].
#
############

import numpy as np

from hermes.utils.time_utils import *


# Print a dictionary (recursively as appropriate).
def print_dict(d):
    print(get_dict_str(d))


# Get a string to display a dictionary (recursively as appropriate).
def get_dict_str(d, level=0):
    indent_root = " " * level
    indent_keys = indent_root + " "
    msg = "%s{\n" % indent_root
    for key, value in d.items():
        msg += "%s %s: " % (indent_keys, key)
        if isinstance(value, dict):
            msg += "\n"
            msg += get_dict_str(
                value, level + 2
            )  # one level for the key indent, one for advancing the level
        else:
            msg += "%s\n" % str(value)
    msg += "%s}\n" % indent_root
    return msg


# Print a variable and its type.
def print_var(var, name=None):
    print(get_var_str(var, name=name))


# Get a string to display a variable and its type.
def get_var_str(var, name=None):
    msg = ""
    if name is not None:
        msg += 'Variable "%s" of ' % name
    msg += "Type %s: " % type(var)
    msg += ""
    processed_var = False
    # Dictionary
    if isinstance(var, dict):
        msg += get_dict_str(var, level=3)
        processed_var = True
    # String
    if isinstance(var, str):
        msg += '"%s"' % var
        processed_var = True
    # Numpy array, if numpy has been imported
    try:
        if isinstance(var, np.ndarray):
            msg += "\n shape: %s" % str(var.shape)
            msg += "\n data type: %s" % str(var.dtype)
            msg += "\n %s" % str(var)
            processed_var = True
    except NameError:
        pass
    # Lists and tuples
    if isinstance(var, (list, tuple)):
        contains_non_numbers = False in [isinstance(x, (int, float)) for x in var]
        if contains_non_numbers:
            msg += "["
            for i, x in enumerate(var):
                msg += "\n %d: %s" % (i, get_var_str(x))
            msg += "\n ]"
        else:
            msg += "%s" % str(var)
        processed_var = True
    # Everything else
    if not processed_var:
        msg += "%s" % str(var)
        processed_var = True
    # Done!
    return msg.strip()


# Format (and optionally print) a timestamped log message.
def format_log_message(
    msg: str,
    *extra_msgs,
    source_tag: str | None = None,
    debug: bool = False,
    warning: bool = False,
    error: bool = False,
    userAction: bool = False,
    **kwargs,
):
    # Add a timestamp.
    timestamp_str = get_time_str(format="%Y-%m-%d %H:%M:%S.%f")
    # timestamp_str = get_time_str(format='%H:%M:%S.%f')
    # Add an indicator of the source.
    source_str = "[%s]" % source_tag
    source_str = source_str.ljust(10)
    # Add an indicator of the importance level.
    levels_str = {
        "normal": "[normal]",
        "debug": "[debug]",
        "warning": "[warn]",
        "error": "[error]",
        "userAction": "[prompt]",
    }
    level_str = levels_str["normal"]
    if debug:
        level_str = levels_str["debug"]
    if warning:
        level_str = levels_str["warning"]
    if error:
        level_str = levels_str["error"]
    if userAction:
        level_str = levels_str["userAction"]
    level_str = level_str.ljust(max([len(x) for x in levels_str.values()]))
    # Add the message prefix to each line of the message.
    msg_prefix = "%s %s %s" % (timestamp_str, source_str, level_str)
    msg = "%s: %s" % (msg_prefix, msg.replace("\n", "\n%s: " % msg_prefix))
    # Mimic how print() would print multiple arguments.
    for extra_msg in extra_msgs:
        msg += " %s" % extra_msg
    # Return the formatted message.
    return msg


# Format, optionally print, and optionally write to file a timestamped log message.
def write_log_message(
    msg: str,
    *extra_msgs,
    source_tag: str | None = None,
    debug: bool = False,
    warning: bool = False,
    error: bool = False,
    userAction: bool = False,
    print_message: bool = True,
    filepath: str | None = None,
    **kwargs,
):
    msg = format_log_message(
        msg,
        *extra_msgs,
        source_tag=source_tag,
        debug=debug,
        warning=warning,
        error=error,
        userAction=userAction,
        print_message=print_message,
        **kwargs,
    )
    if filepath is not None:
        # Note that it seems multiple file handles can be opened at the same time,
        #  so it shouldn't get an error in the case of multiple threads calling
        #  this method simultaneously.  But surround with try/except just in case.
        try:
            fout = open(filepath, "a")
            fout.write(msg)
            fout.write("\n")
            fout.close()
        except:
            print(
                "\n\nWARNING: error trying to write a log message to the following file: ",
                filepath,
                "\n\n",
            )
    return msg


######################
###### PRINTING ######
######################


def log_status(source_tag, msg, *extra_msgs, **kwargs) -> str:
    return format_log_message(msg, *extra_msgs, source_tag=source_tag, **kwargs)


def log_debug(source_tag, msg, *extra_msgs, **kwargs) -> str:
    return format_log_message(
        msg, *extra_msgs, source_tag=source_tag, debug=True, **kwargs
    )


def log_error(source_tag, msg, *extra_msgs, **kwargs) -> str:
    return format_log_message(
        msg, *extra_msgs, source_tag=source_tag, error=True, **kwargs
    )


def log_warn(source_tag, msg, *extra_msgs, **kwargs) -> str:
    return format_log_message(
        msg, *extra_msgs, source_tag=source_tag, warning=True, **kwargs
    )



############
#
# Copyright (c) 2022 MIT CSAIL and Joseph DelPreto
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# See https://action-net.csail.mit.edu for more usage information.
# Created 2021-2022 for the MIT ActionNet project by Joseph DelPreto [https://josephdelpreto.com].
#
############

import pyperclip


# A helper to insert new entries into the end of an array,
#  shifting back and deleting earlier entries.
# Do it in-place instead of using append() and then trimming,
#  to avoid making inefficient copies.
# Assumes arr and new_entries are 1D arrays.
def add_to_rolling_array(arr, new_entries):
    # Determine start/end indexes of the new entries such that it
    #  can fit in the array to update.
    new_entries_end_index = new_entries.shape[0] - 1
    new_entries_start_index = new_entries_end_index - arr.shape[0] + 1
    new_entries_start_index = max(0, new_entries_start_index)
    # Shift original entries back and add the new entries.
    num_new_elements = new_entries_end_index - new_entries_start_index + 1
    arr[0:-num_new_elements] = arr[num_new_elements:]
    arr[-num_new_elements:] = new_entries[
        new_entries_start_index : (new_entries_end_index + 1)
    ]
    return arr


# Get a matrix as a string that can be printed/copied.
def get_matrix_str(matrix, column_delim="\t", row_delim="\n"):
    return row_delim.join(
        [
            column_delim.join([str(x) for x in matrix[row_index, :]])
            for row_index in range(matrix.shape[0])
        ]
    )


# Copy a matrix as a string so it can be pasted into Excel or a similar program.
def copy_matrix(matrix, column_delim="\t", row_delim="\n"):
    pyperclip.copy(
        get_matrix_str(matrix, column_delim=column_delim, row_delim=row_delim)
    )



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from hermes.utils.types import LoggingSpec
from hermes.utils.zmq_utils import PORT_FRONTEND, PORT_KILL, PORT_SYNC_HOST

from hermes.base.nodes.consumer import Consumer


class DataLogger(Consumer):
    """A Node to centrally log streaming data of all specified modalities to one or more files.

    May be redundant because each Node has a local Storage component that manages safe disk flushing of the acquired data.
    Producer and Pipeline specifications mapping is passed to the class,
    and the data they stream are written to disk periodically and/or at the end.
    """

    @classmethod
    def _log_source_tag(cls) -> str:
        return "logger"

    def __init__(
        self,
        host_ip: str,
        stream_specs: list[dict],
        logging_spec: LoggingSpec,
        port_sub: str = PORT_FRONTEND,
        port_sync: str = PORT_SYNC_HOST,
        port_killsig: str = PORT_KILL,
        log_history_filepath: str | None = None,
        **_,
    ):
        """Constructor of the centralized Storage Node.

        Args:
            host_ip (str): IP address of the local master Broker.
            stream_specs (list[dict]): List of mappings of user-configured incoming modalities.
            logging_spec (LoggingSpec): Mapping of Storage object parameters to user-defined configuration values.
            port_sub (str, optional): Local port to subscribe to for incoming relayed data from the local master Broker. Defaults to `PORT_FRONTEND`.
            port_sync (str, optional): Local port to listen to for local master Broker's startup coordination. Defaults to `PORT_SYNC_HOST`.
            port_killsig (str, optional): Local port to listen to for local master Broker's termination signal. Defaults to `PORT_KILL`.
            log_history_filepath (str | None, optional): File path to the system log file. Defaults to `None`.
        """
        super().__init__(
            host_ip=host_ip,
            stream_in_specs=stream_specs,
            logging_spec=logging_spec,
            port_sub=port_sub,
            port_sync=port_sync,
            port_killsig=port_killsig,
            log_history_filepath=log_history_filepath,
        )

    def _cleanup(self):
        super()._cleanup()



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from hermes.utils.zmq_utils import PORT_FRONTEND, PORT_KILL, PORT_SYNC_HOST
from hermes.utils.types import LoggingSpec

from hermes.base.nodes.consumer import Consumer


class DummyConsumer(Consumer):
    """A Node showcasing the Consumer behavior, consumeing external data relayed to it by the Broker."""

    @classmethod
    def _log_source_tag(cls) -> str:
        return "dummy-consumer"

    def __init__(
        self,
        host_ip: str,
        stream_in_specs: list[dict],
        logging_spec: LoggingSpec,
        port_sub: str = PORT_FRONTEND,
        port_sync: str = PORT_SYNC_HOST,
        port_killsig: str = PORT_KILL,
        **_,
    ):
        """Constructor of the DummyConsumer Node.

        Args:
            host_ip (str): IP address of the local master Broker.
            stream_in_specs (list[dict]): List of mappings of user-configured incoming modalities.
            logging_spec (LoggingSpec): Specification of what and how to store.
            port_sub (str, optional): Local port to subscribe to for incoming relayed data from the local master Broker. Defaults to `PORT_FRONTEND`.
            port_sync (str, optional): Local port to listen to for local master Broker's startup coordination. Defaults to `PORT_SYNC_HOST`.
            port_killsig (str, optional): Local port to listen to for local master Broker's termination signal. Defaults to `PORT_KILL`.
        """
        super().__init__(
            host_ip=host_ip,
            stream_in_specs=stream_in_specs,
            logging_spec=logging_spec,
            port_sub=port_sub,
            port_sync=port_sync,
            port_killsig=port_killsig,
        )

    def _cleanup(self):
        super()._cleanup()
